"""merge b097e47ba829 with cdd0dc9d9870

Revision ID: 079a74c15e8b
Revises: b097e47ba829, cdd0dc9d9870
Create Date: 2021-05-30 20:49:19.039280

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '079a74c15e8b'
down_revision = ('b097e47ba829', 'cdd0dc9d9870')
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
